package limax.auany;

import java.util.Arrays;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;

import javax.management.Attribute;
import javax.management.AttributeList;
import javax.management.AttributeNotFoundException;
import javax.management.DynamicMBean;
import javax.management.InvalidAttributeValueException;
import javax.management.MBeanAttributeInfo;
import javax.management.MBeanException;
import javax.management.MBeanInfo;
import javax.management.MBeanOperationInfo;
import javax.management.MBeanParameterInfo;
import javax.management.ReflectionException;

import limax.auany.switcherauany.CheckProviderKey;
import limax.auany.switcherauany.PayIdsNotify;
import limax.defines.ErrorCodes;
import limax.net.Transport;
import limax.util.MBeans;
import limax.util.Trace;

final class ProviderKeyStore {
	private final static Map<Integer, Set<Transport>> transportmap = new ConcurrentHashMap<>();
	private final static Map<Integer, String> keymap = new ConcurrentHashMap<>();

	static void insertKey(int pvid, String key) {
		keymap.put(pvid, key);
	}

	static int checkKey(int pvid, String key) {
		final String s = keymap.get(pvid);
		if (null == s)
			return ErrorCodes.AUANY_CHECK_PROVIDER_KEY_UNKNOWN_PVID;
		if (s.equals(key))
			return ErrorCodes.SUCCEED;
		else
			return ErrorCodes.AUANY_CHECK_PROVIDER_KEY_BAD_KEY;
	}

	private static void bind(int payid, Transport transport) {
		Set<Transport> set = Collections.newSetFromMap(new ConcurrentHashMap<Transport, Boolean>());
		Set<Transport> oldset = transportmap.putIfAbsent(payid, set);
		if (oldset != null)
			set = oldset;
		set.add(transport);
		PayDelivery.flushQueue(payid, transport);
	}

	static void unbind(Transport transport) {
		transportmap.values().forEach(set -> set.remove(transport));
	}

	static Transport getTransportByPayId(int payid) {
		Set<Transport> transports = transportmap.get(payid);
		if (transports == null)
			return null;
		return transports.stream().findAny().orElse(null);
	}

	static void process(PayIdsNotify notify) {
		notify.payids.forEach(payid -> bind(payid, notify.getTransport()));
	}

	static void process(CheckProviderKey rpc) {
		final limax.switcherauany.CheckProviderKeyArg arg = rpc.getArgument();
		if ((rpc.getResult().errorcodes = checkKey(arg.pvid, arg.pvkey)) == ErrorCodes.SUCCEED && arg.payid != 0)
			bind(arg.payid, rpc.getTransport());
		if (Trace.isInfoEnabled())
			Trace.info("CheckProviderKey pvid=" + arg.pvid + ",payid=" + arg.payid + ",errorcodes="
					+ rpc.getResult().errorcodes);
	}

	static {
		MBeans.register(MBeans.root(), new MBeanOperators(), "limax.auany:type=ProviderKeyStore");
	}

	private final static class MBeanOperators implements DynamicMBean {

		@Override
		public Object getAttribute(String attribute)
				throws AttributeNotFoundException, MBeanException, ReflectionException {
			switch (attribute) {
			case "keys": {
				return keymap;
			}
			default:
				break;
			}
			return null;
		}

		@Override
		public void setAttribute(Attribute attribute)
				throws AttributeNotFoundException, InvalidAttributeValueException, MBeanException, ReflectionException {
		}

		@Override
		public AttributeList getAttributes(String[] attributes) {
			return new AttributeList(Arrays.asList(new Attribute("keys", keymap)));
		}

		@Override
		public AttributeList setAttributes(AttributeList attributes) {
			return null;
		}

		@Override
		public Object invoke(String actionName, Object[] params, String[] signature)
				throws MBeanException, ReflectionException {
			switch (actionName) {
			case "providerid": {
				insertKey((int) params[0], params[1].toString());
				return 0;
			}
			default:
				break;
			}
			return null;
		}

		@Override
		public MBeanInfo getMBeanInfo() {
			MBeanAttributeInfo[] attrinfos = {
					new MBeanAttributeInfo("keys", Map.class.getName(), "keys", true, false, false) };
			MBeanOperationInfo[] operinfos = {
					new MBeanOperationInfo("setOrInsertKey", "update the old key value, or insert a new key",
							Arrays.asList(new MBeanParameterInfo("providerid", int.class.getName(), "provider id"),
									new MBeanParameterInfo("key", String.class.getName(), "provider key"))
							.toArray(new MBeanParameterInfo[0]), String.class.getName(), MBeanOperationInfo.ACTION) };
			return new MBeanInfo(ProviderKeyStore.class.getName(), "provider keys", attrinfos, null, operinfos, null);
		}

	}
}
