package limax.codec;

import java.io.IOException;
import java.lang.reflect.Array;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.atomic.AtomicLong;

public final class JSONEncoder {
	private interface Action {
		void apply(Set<Object> l, Object v, Appendable a) throws IOException;
	}

	private final static Map<Class<?>, Action> actions = new ConcurrentHashMap<>();
	private final static Action stringAction = (l, v, a) -> {
		a.append('"');
		String s = v.toString();
		for (int i = 0, n = s.length(); i < n; i++) {
			char c = s.charAt(i);
			switch (c) {
			case '"':
				a.append("\\\"");
				break;
			case '\\':
				a.append("\\\\");
				break;
			case '\b':
				a.append("\\b");
				break;
			case '\f':
				a.append("\\f");
				break;
			case '\n':
				a.append("\\n");
				break;
			case '\r':
				a.append("\\r");
				break;
			case '\t':
				a.append("\\t");
				break;
			default:
				a.append(c);
			}
		}
		a.append('"');
	};
	private final static Action arrayAction = (l, v, a) -> {
		String comma = "";
		a.append('[');
		for (int i = 0, n = Array.getLength(v); i < n; i++) {
			a.append(comma);
			encode(l, Array.get(v, i), a);
			comma = ",";
		}
		a.append(']');
	};
	private final static Action collectionAction = (l, v, a) -> {
		String comma = "";
		a.append('[');
		for (Object i : (Collection<?>) v) {
			a.append(comma);
			encode(l, i, a);
			comma = ",";
		}
		a.append(']');
	};

	static {
		Action numberAction = (l, v, a) -> a.append(v.toString());
		Action booleanAction = (l, v, a) -> a.append(Boolean.toString((boolean) v));
		actions.put(byte.class, numberAction);
		actions.put(short.class, numberAction);
		actions.put(int.class, numberAction);
		actions.put(long.class, numberAction);
		actions.put(float.class, numberAction);
		actions.put(double.class, numberAction);
		actions.put(Byte.class, numberAction);
		actions.put(Short.class, numberAction);
		actions.put(Integer.class, numberAction);
		actions.put(Long.class, numberAction);
		actions.put(Float.class, numberAction);
		actions.put(Double.class, numberAction);
		actions.put(AtomicInteger.class, numberAction);
		actions.put(AtomicLong.class, numberAction);
		actions.put(boolean.class, booleanAction);
		actions.put(Boolean.class, booleanAction);
		actions.put(String.class, stringAction);
	}

	private static class FieldAction implements Action {
		private final Field field;
		private final Action action;

		FieldAction(Field field, Action action) {
			this.field = field;
			this.action = action;
		}

		@Override
		public void apply(Set<Object> l, Object v, Appendable a) throws IOException {
			Object o = null;
			try {
				o = field.get(v);
			} catch (Exception e) {
			}
			if (o instanceof JSONSerializable && !l.add(o))
				throw new IOException("JSONEncoder loop detected. object = " + o);
			stringAction.apply(l, field.getName(), a);
			a.append(':');
			if (o != null)
				action.apply(l, o, a);
			else
				a.append("null");
		}
	}

	private static boolean findSuperClass(Class<?> clazz, Class<?> superclass) {
		if (clazz == superclass)
			return true;
		for (Class<?> c : clazz.getInterfaces())
			if (findSuperClass(c, superclass))
				return true;
		Class<?> c = clazz.getSuperclass();
		return c == null ? false : findSuperClass(c, superclass);
	}

	private static Action compile(Class<?> c) throws IOException {
		Action action = actions.get(c);
		if (action != null)
			return action;
		if (c.isArray())
			return arrayAction;
		if (findSuperClass(c, Collection.class))
			return collectionAction;
		if (!findSuperClass(c, JSONSerializable.class))
			throw new IOException("JSONEncoder encounter unrecognized type " + c.getName());
		FieldAction[] fieldActions = Arrays.stream(c.getDeclaredFields()).peek(field -> field.setAccessible(true))
				.filter(field -> (field.getModifiers() & (Modifier.TRANSIENT | Modifier.STATIC)) == 0)
				.map(field -> new FieldAction(field, (l, v, a) -> encode(l, v, a))).toArray(FieldAction[]::new);
		actions.putIfAbsent(c, action = (l, v, a) -> {
			String comma = "";
			a.append('{');
			for (FieldAction fieldAction : fieldActions) {
				a.append(comma);
				fieldAction.apply(l, v, a);
				comma = ",";
			}
			a.append('}');
		});
		return action;
	}

	private static void encode(Set<Object> l, Object v, Appendable a) throws IOException {
		if (v != null)
			compile(v.getClass()).apply(l, v, a);
		else
			a.append("null");
	}

	public static void encode(Object obj, Appendable a) throws CodecException {
		try {
			encode(new HashSet<Object>(), obj, a);
		} catch (IOException e) {
			throw new CodecException(e);
		}
		if (a instanceof Source)
			((Source) a).flush();
	}
}
