package limax.endpoint;

import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.Executor;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;

import limax.codec.CharSink;
import limax.codec.JSON;
import limax.codec.JSONDecoder;
import limax.defines.ErrorCodes;
import limax.defines.ErrorSource;
import limax.endpoint.auanyviews.Service;
import limax.endpoint.auanyviews.ServiceResult;
import limax.net.Config;
import limax.net.Engine;
import limax.net.Manager;
import limax.net.Transport;
import limax.util.HttpClient;

public final class AuanyService {
	public final static int providerId = 1;

	public interface ResultPay {
		void accept(int errorSource, int errorCode, String order);
	}

	public interface ResultCredential {
		void accept(int errorSource, int errorCode, String credential);
	}

	private final static AtomicInteger snGenerator = new AtomicInteger();
	private final static Map<Integer, AuanyService> map = new HashMap<Integer, AuanyService>();

	private final int sn;
	private final Executor executor;
	private final Object onresult;
	private final Future<?> future;

	private static AuanyService removeService(int sn) {
		synchronized (map) {
			AuanyService service = map.remove(sn);
			if (service != null)
				service.future.cancel(false);
			return service;
		}
	}

	private AuanyService(final Object onresult, long timeout, Executor executor) {
		this.sn = snGenerator.getAndIncrement();
		this.executor = executor;
		this.onresult = onresult;
		synchronized (map) {
			this.future = Engine.getProtocolScheduler().schedule(new Runnable() {
				@Override
				public void run() {
					AuanyService service = removeService(sn);
					if (service != null)
						service.onError(ErrorCodes.ENDPOINT_AUANY_SERVICE_CLIENT_TIMEOUT);
				}
			}, timeout, TimeUnit.MILLISECONDS);
			map.put(this.sn, this);
		}
	}

	private void onError(int errorCode) {
		if (onresult instanceof ResultCredential)
			onResultCredit(ErrorSource.ENDPOINT, errorCode, null);
		else if (onresult instanceof ResultPay)
			onResultPay(ErrorSource.ENDPOINT, errorCode, null);
	}

	private void onResultCredit(final int errorSource, final int errorCode, final String credential) {
		executor.execute(new Runnable() {
			@Override
			public void run() {
				((ResultCredential) onresult).accept(errorSource, errorCode, credential);
			}
		});
	}

	private void onResultPay(final int errorSource, final int errorCode, final String order) {
		executor.execute(new Runnable() {
			@Override
			public void run() {
				((ResultPay) onresult).accept(errorSource, errorCode, order);
			}
		});
	}

	static void onResultViewOpen(ServiceResult view) {
		view.registerListener("resultCredential", new ViewChangedListener() {
			@Override
			public void onViewChanged(ViewChangedEvent e) {
				limax.auanyviews.ResultCredential r = (limax.auanyviews.ResultCredential) e.getValue();
				AuanyService service = removeService(r.sn);
				if (service != null)
					service.onResultCredit(r.errorSource, r.errorCode, r.credential);
			}
		});
		view.registerListener("resultPay", new ViewChangedListener() {
			@Override
			public void onViewChanged(ViewChangedEvent e) {
				limax.auanyviews.ResultPay r = (limax.auanyviews.ResultPay) e.getValue();
				AuanyService service = removeService(r.sn);
				if (service != null)
					service.onResultPay(r.errorSource, r.errorCode, r.order);
			}
		});
	}

	static void cleanup() {
		synchronized (map) {
			for (AuanyService service : map.values()) {
				service.future.cancel(false);
				service.onError(ErrorCodes.ENDPOINT_AUANY_SERVICE_ENGINE_CLOSE);
			}
			map.clear();
		}
	}

	private static Executor createExecutor(final EndpointManager manager) {
		return new Executor() {
			@Override
			public void execute(Runnable command) {
				((EndpointManagerImpl) manager).dispatch(command, manager.getTransport());
			}
		};
	}

	private static class CredentialContext {
		private int errorSource = ErrorSource.ENDPOINT;
		private int errorCode = ErrorCodes.ENDPOINT_AUANY_SERVICE_CLIENT_TIMEOUT;
		private String credential;
		private Manager manager;
		private final String httpHost;
		private final int httpPort;
		private final int appId;
		private final String nonce;
		private final String username;
		private final String token;
		private final String platflag;
		private final long timeout;
		private final ResultCredential resultCredit = new ResultCredential() {

			@Override
			public void accept(int errorSource, int errorCode, String credential) {
				synchronized (CredentialContext.this) {
					CredentialContext.this.errorSource = errorSource;
					CredentialContext.this.errorCode = errorCode;
					CredentialContext.this.credential = credential;
					CredentialContext.this.notify();
				}
			}
		};

		private CredentialContext(String httpHost, int httpPort, int appId, String nonce, String username, String token,
				String platflag, long timeout) {
			this.httpHost = httpHost;
			this.httpPort = httpPort;
			this.appId = appId;
			this.nonce = nonce;
			this.username = username;
			this.token = token;
			this.platflag = platflag;
			this.timeout = timeout;
		}

		private CredentialContext(String httpHost, int httpPort, int appId, String nonce, long timeout) {
			this(httpHost, httpPort, appId, nonce, null, null, null, timeout);
		}

		public void execute(ResultCredential r) {
			try {
				long starttime = System.currentTimeMillis();
				JSONDecoder decoder = new JSONDecoder();
				new HttpClient("http://" + httpHost + ":" + httpPort + "/invite?native=" + appId, timeout / 2, 4096,
						null, false).transfer(new CharSink(decoder));
				JSON json = decoder.get();
				JSON switcher = json.get("switcher");
				EndpointConfig config = Endpoint.createLoginConfigBuilder(switcher.get("host").toString(),
						switcher.get("port").intValue(), json.get("code").toString(), "", "invite").build();
				long remain = timeout - (System.currentTimeMillis() - starttime);
				if (remain > 0)
					synchronized (this) {
						Endpoint.start(config, new EndpointListener() {
							@Override
							public void onAbort(Transport transport) throws Exception {
								synchronized (CredentialContext.this) {
									CredentialContext.this.notify();
								}
							}

							@Override
							public void onManagerInitialized(Manager manager, Config config) {
								CredentialContext.this.manager = manager;
							}

							@Override
							public void onManagerUninitialized(Manager manager) {
							}

							@Override
							public void onTransportAdded(Transport transport) throws Exception {
								if (username == null)
									AuanyService.derive(nonce, timeout, resultCredit);
								else
									AuanyService.bind(nonce, username, token, platflag, timeout, resultCredit);
							}

							@Override
							public void onTransportRemoved(Transport transport) throws Exception {
							}

							@Override
							public void onSocketConnected() {
							}

							@Override
							public void onKeyExchangeDone() {
							}

							@Override
							public void onKeepAlived(int ms) {
							}

							@Override
							public void onErrorOccured(int source, int code, Throwable exception) {
								synchronized (CredentialContext.this) {
									errorSource = source;
									errorCode = code;
									CredentialContext.this.notify();
								}
							}
						});
						wait(remain);
					}
				manager.close();
			} catch (Exception e) {
			}
			r.accept(errorSource, errorCode, credential);
		}
	}

	public static void bind(String nonce, String username, String token, String platflag, long timeout,
			ResultCredential onresult, EndpointManager manager) throws Exception {
		Service.getInstance(manager).Bind(new AuanyService(onresult, timeout, createExecutor(manager)).sn, nonce,
				username, token, platflag);
	}

	public static void bind(String nonce, String username, String token, String platflag, long timeout,
			ResultCredential onresult) throws Exception {
		bind(nonce, username, token, platflag, timeout, onresult, Endpoint.getDefaultEndpointManager());
	}

	public static void bind(String httpHost, int httpPort, int appId, String nonce, String username, String token,
			String platflag, long timeout, ResultCredential onresult) {
		new CredentialContext(httpHost, httpPort, appId, nonce, username, token, platflag, timeout).execute(onresult);
	}

	public static void derive(String nonce, long timeout, ResultCredential onresult, EndpointManager manager)
			throws Exception {
		Service.getInstance(manager).Derive(new AuanyService(onresult, timeout, createExecutor(manager)).sn, nonce);
	}

	public static void derive(String nonce, long timeout, ResultCredential onresult) throws Exception {
		derive(nonce, timeout, onresult, Endpoint.getDefaultEndpointManager());
	}

	public static void derive(String httpHost, int httpPort, int appId, String nonce, long timeout,
			ResultCredential onresult) {
		new CredentialContext(httpHost, httpPort, appId, nonce, timeout).execute(onresult);
	}

	public static void pay(int gateway, int payid, int product, int price, int quantity, String receipt, long timeout,
			ResultPay onresult, EndpointManager manager) throws Exception {
		Service.getInstance(manager).Pay(new AuanyService(onresult, timeout, createExecutor(manager)).sn, gateway,
				payid, product, price, quantity, receipt);
	}

	public static void pay(int gateway, int payid, int product, int price, int quantity, String receipt, long timeout,
			ResultPay onresult) throws Exception {
		pay(gateway, payid, product, price, quantity, receipt, timeout, onresult, Endpoint.getDefaultEndpointManager());
	}

}
