package limax.provider;

import java.io.IOException;
import java.util.Collections;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.concurrent.locks.LockSupport;

import limax.codec.CodecException;
import limax.codec.Octets;
import limax.codec.OctetsStream;
import limax.defines.ErrorCodes;
import limax.defines.ErrorSource;
import limax.net.AbstractRpcContext;
import limax.net.Config;
import limax.net.Driver;
import limax.net.Engine;
import limax.net.Listener;
import limax.net.Manager;
import limax.net.ManagerConfig;
import limax.net.SizePolicyException;
import limax.net.Skeleton;
import limax.net.SupportDispatch;
import limax.net.SupportTypedDataBroadcast;
import limax.net.Transport;
import limax.net.WebSocketProtocol;
import limax.provider.providerendpoint.SendControlToServer;
import limax.provider.providerendpoint.SyncViewToClients;
import limax.provider.switcherprovider.Bind;
import limax.provider.switcherprovider.BindResult;
import limax.provider.switcherprovider.Broadcast;
import limax.provider.switcherprovider.Dispatch;
import limax.provider.switcherprovider.LinkBroken;
import limax.provider.switcherprovider.OnlineAnnounce;
import limax.provider.switcherprovider.Pay;
import limax.provider.switcherprovider.PayAck;
import limax.provider.switcherprovider.UnBind;
import limax.provider.switcherprovider.Unicast;
import limax.util.Dispatcher;
import limax.util.Enable;
import limax.util.Trace;

class ProviderManagerImpl extends AbstractRpcContext
		implements ProviderManager, SupportTypedDataBroadcast, SupportDispatch {
	private final ProviderListener listener;
	private final ProviderManagerConfig config;
	private final ProviderSwitcherExchanger exchanger;
	private final Set<Manager> managers = Collections.newSetFromMap(new ConcurrentHashMap<Manager, Boolean>());
	private final Manager wrapper;
	private final Dispatcher dispatcher;
	private final ViewContextImpl viewcontext;
	private volatile boolean closed = false;
	private volatile boolean hasbind = false;
	private final Map<Long, ProviderTransportImpl> transports = new ConcurrentHashMap<>();
	private final AtomicInteger sessionCounter = new AtomicInteger();
	private volatile Thread closingThread;

	ProviderManagerImpl(ProviderManagerConfig config, ProviderListener listener, Manager wrapper) throws Exception {
		this.exchanger = new ProviderSwitcherExchanger(this);
		this.listener = listener;
		this.config = config;
		this.wrapper = wrapper;
		this.dispatcher = new Dispatcher(Engine.getProtocolExecutor());
		this.viewcontext = new ViewContextImpl(config, config.getViewManagerClassName()) {
			@Override
			void syncViewToClients(SyncViewToClients protocol)
					throws InstantiationException, ClassCastException, SizePolicyException, CodecException {
				toswitcher.syncViewToClients(protocol);
			}

			@Override
			ViewSession getViewSession(long sessionid) {
				ProviderTransportImpl transport = transports.get(sessionid);
				return transport == null ? null : transport.getViewSession();
			}
		};
		checkSupported();
		try {
			for (ManagerConfig cfg : config.getManagerConfigs())
				managers.add(exchanger.add(cfg));
		} catch (Exception e) {
			managers.forEach(Manager::close);
			throw e;
		}
		listener.onManagerInitialized(this, config);
	}

	boolean isStateless() {
		return config.getSessionTimeout() > 0;
	}

	private void checkSupported() {
		if (Enable.True == config.getAllowUseScript() && !viewcontext.isScriptSupported())
			throw new RuntimeException("config allowUseScript = true but viewcontext.isScriptSupported = false");
		if (Enable.True == config.getAllowUseVariant() && !viewcontext.isVariantSupported())
			throw new RuntimeException("config allowUseVariant = true but viewcontext.isVariantSupported = false");
		if (Trace.isWarnEnabled()) {
			if (Enable.False == config.getAllowUseScript() && viewcontext.isScriptSupported())
				Trace.warn("config allowUseScript = false but viewcontext.isScriptSupported = true");
			if (Enable.False == config.getAllowUseVariant() && viewcontext.isVariantSupported())
				Trace.warn("config allowUseVariant = false but viewcontext.isVariantSupported = true");
		}
	}

	void scheduleSessionTask(long sid, Runnable task) {
		Engine.getApplicationExecutor().execute(sid, task);
	}

	static {
		Engine.registerDriver(new Driver() {
			@Override
			public Class<? extends Config> getConfigClass() {
				return ProviderManagerConfig.class;
			}

			@Override
			public Class<? extends Listener> getListenerClass() {
				return ProviderListener.class;
			}

			@Override
			public Manager newInstance(Config config, Listener listener, Manager wrapper) throws Exception {
				return new ProviderManagerImpl((ProviderManagerConfig) config, (ProviderListener) listener, wrapper);
			}
		});
	}

	@Override
	public void close() {
		closed = true;
		if (Engine.remove(this))
			return;
		closingThread = Thread.currentThread();
		exchanger.close();
		managers.forEach(Manager::close);
		dispatcher.await();
		viewcontext.close();
		while (sessionCounter.get() != 0)
			LockSupport.park(sessionCounter);
		listener.onManagerUninitialized(this);
	}

	@Override
	public void close(long sessionid) {
		exchanger.kick(sessionid, ErrorCodes.PROVIDER_KICK_SESSION);
		scheduleSessionTask(sessionid, () -> removeTransport(sessionid));
	}

	@Override
	public void close(Transport transport) {
		close(((ProviderTransportImpl) transport).getSessionId());
	}

	@Override
	public ProviderListener getListener() {
		return listener;
	}

	@Override
	public Config getConfig() {
		return config;
	}

	int getPvid() {
		return config.getProviderId();
	}

	@Override
	public String toString() {
		return config.getName();
	}

	void statelessRemoveTransport(long sid) {
		ProviderTransportImpl transport = transports.remove(sid);
		if (null == transport)
			return;
		if (null != transport.futureForStateless)
			transport.futureForStateless.cancel(false);
		exchanger.removeSession(transport.getRelativeTransport(), sid);
		transport.getViewSession().close(sid, false);
	}

	private ProviderTransportImpl _removeTransport(long sid) {
		ProviderTransportImpl transport = transports.remove(sid);
		if (null == transport)
			return null;
		exchanger.removeSession(transport.getRelativeTransport(), sid);
		return transport;
	}

	void removeTransport(long sid) {
		ProviderTransportImpl transport = _removeTransport(sid);
		if (transport == null)
			return;
		Engine.getProtocolExecutor().execute(sid, () -> {
			transport.getViewSession().close(sid, true);
			scheduleSessionTask(sid, () -> {
				try {
					listener.onTransportRemoved(transport);
				} catch (Throwable e) {
					if (Trace.isErrorEnabled())
						Trace.error(this + " removeTransport = " + transport, e);
				} finally {
					if (sessionCounter.decrementAndGet() == 0)
						LockSupport.unpark(closingThread);
				}
			});
		});
	}

	private void addTransport1(OnlineAnnounce oa) {
		ProviderTransportImpl transport = transports.get(oa.sid);
		if (null == transport) {
			addTransport3(oa);
		} else {
			try {
				listener.onTransportDuplicate(transport);
			} catch (Throwable e) {
				close(transport);
			}
			scheduleSessionTask(oa.sid, () -> addTransport2(oa));
		}
	}

	private void addTransport2(OnlineAnnounce oa) {
		if (!transports.containsKey(oa.sid)) {
			Engine.getProtocolExecutor().execute(oa.sid, () -> scheduleSessionTask(oa.sid, () -> addTransport3(oa)));
		} else {
			limax.provider.switcherendpoint.OnlineAnnounce protocol = new limax.provider.switcherendpoint.OnlineAnnounce();
			protocol.sid = oa.connectionid;
			protocol.errorSource = ErrorSource.LIMAX;
			protocol.errorCode = ErrorCodes.PROVIDER_SESSION_LOGINED;
			try {
				protocol.send(oa.getTransport());
			} catch (Throwable e) {
				if (Trace.isInfoEnabled())
					Trace.info(this + " addTransport2 = " + oa.getTransport(), e);
			}
		}
	}

	private void addTransport3(OnlineAnnounce oa) {
		if (!exchanger.addSession(oa.getTransport(), oa.sid))
			return;
		ProviderTransportImpl transport = new ProviderTransportImpl(this, oa);
		transports.put(oa.sid, transport);
		limax.provider.switcherendpoint.OnlineAnnounce protocol = new limax.provider.switcherendpoint.OnlineAnnounce();
		protocol.sid = oa.connectionid;
		protocol.errorSource = ErrorSource.LIMAX;
		try {
			listener.onTransportAdded(transport);
			sessionCounter.incrementAndGet();
		} catch (Throwable e) {
			if (Trace.isErrorEnabled())
				Trace.error(this + " addTransport3 = " + transport, e);
			_removeTransport(oa.sid);
			protocol.errorCode = ErrorCodes.PROVIDER_ADD_TRANSPORT_EXCEPTION;
			try {
				protocol.send(oa.getTransport());
			} catch (Throwable e2) {
				if (Trace.isInfoEnabled())
					Trace.info(this + " addTransport3 = " + oa.getTransport(), e2);
			}
			return;
		}
		try {
			viewcontext.onSessionLogin(oa.sid, transport.getViewSession());
			protocol.errorCode = ErrorCodes.SUCCEED;
			protocol.flags = oa.flags;
			protocol.send(oa.getTransport());
		} catch (Throwable e) {
			if (Trace.isErrorEnabled())
				Trace.error(this + " addTransport3 = " + transport, e);
			removeTransport(oa.sid);
		}
	}

	private void statelessResetSessionTimeout(ProviderTransportImpl transport, long sessionid) {
		if (null != transport.futureForStateless)
			transport.futureForStateless.cancel(false);
		transport.futureForStateless = Engine.getProtocolScheduler().schedule(
				() -> scheduleSessionTask(sessionid, () -> statelessRemoveTransport(sessionid)),
				config.getSessionTimeout(), TimeUnit.MILLISECONDS);
	}

	private void statelessAddTransport(long sessionid, Transport from) {
		ProviderTransportImpl transport = transports.get(sessionid);
		if (null != transport) {
			statelessResetSessionTimeout(transport, sessionid);
			return;
		}
		if (!exchanger.addSession(from, sessionid))
			return;
		transport = new ProviderTransportImpl(this, sessionid, from);
		transports.put(sessionid, transport);
		try {
			viewcontext.onSessionLogin(sessionid, transport.getViewSession());
			statelessResetSessionTimeout(transport, sessionid);
			sessionCounter.incrementAndGet();
		} catch (Throwable e) {
			if (Trace.isErrorEnabled())
				Trace.error(this + " addTransport = " + transport, e);
			statelessRemoveTransport(sessionid);
		}
	}

	@Override
	public void dispatch(Runnable r, Object hit) {
		if (r instanceof Dispatch) {
			final Dispatch dp = (Dispatch) r;
			providerdispatcher.dispatchSessionProtocol(dp.sid, dp.ptype, dp.pdata);
		} else if (r instanceof SendControlToServer)
			providerdispatcher.onViewControl((SendControlToServer) r);
		else
			dispatcher.execute(r, hit);
	}

	@Override
	public void broadcast(int ptype, Octets pdata)
			throws InstantiationException, SizePolicyException, CodecException, ClassCastException {
		toswitcher.broadcastToSwitcher(ptype | (config.getProviderId() << 8), pdata);
	}

	void send(long sid, int ptype, Octets pdata)
			throws InstantiationException, SizePolicyException, CodecException, ClassCastException {
		toswitcher.unicastToSwitcher(sid, ptype | (config.getProviderId() << 8), pdata);
	}

	@Override
	public boolean isListening() {
		return hasbind;
	}

	private Bind createBindProtocol() {
		final Bind bind = new Bind();
		bind.pvid = config.getProviderId();
		bind.pvkey = config.getProviderKey();
		bind.payid = config.getPayId();
		bind.pinfos.putAll(config.getProviderProtocolInfos());
		bind.providerStatus = 0;
		if (viewcontext.isVariantEnabled()) {
			bind.variantdefines = viewcontext.getVariantDefines();
			bind.providerStatus |= Bind.PS_VARIANT_ENABLED;
		}
		if (viewcontext.isScriptEnabled()) {
			bind.scriptdefines = viewcontext.getScriptDefines();
			bind.providerStatus |= Bind.PS_SCRIPT_ENABLED;
		}
		if (viewcontext.isScriptSupported())
			bind.providerStatus |= Bind.PS_SCRIPT_SUPPORTED;
		if (viewcontext.isVariantSupported())
			bind.providerStatus |= Bind.PS_VARIANT_SUPPORTED;
		if (isStateless())
			bind.providerStatus |= Bind.PS_STATELESS;
		return bind;
	}

	synchronized void tryBind(Transport transport) throws Exception {
		if (!hasbind || closed)
			return;
		createBindProtocol().send(transport);
	}

	@Override
	public synchronized void openListen() throws IOException {
		if (closed)
			throw new IOException("Provider closed");
		if (1 == config.getManagerConfigs().size())
			DirectDispatcher.getInstance().register(config.getProviderId(), providerdispatcher);
		try {
			exchanger.send(createBindProtocol());
		} catch (CodecException e) {
			throw new IOException(e);
		}
		hasbind = true;
	}

	@Override
	public synchronized void closeListen() throws IOException {
		if (closed)
			throw new IOException("Provider closed");
		DirectDispatcher.getInstance().unregister(config.getProviderId());
		toswitcher = createDefaultToSwitcher();
		try {
			exchanger.send(new UnBind());
		} catch (CodecException e) {
			throw new IOException(e);
		}
		hasbind = false;
	}

	final void process(BindResult p) {
		if (ErrorCodes.SUCCEED != p.errorcode) {
			Trace.fatal("provider openlisten failed, pvid = " + p.pvid + " errorcode = " + p.errorcode);
			System.exit(0);
		} else if (Trace.isInfoEnabled())
			Trace.info("provider had bind success! pvid = " + p.pvid);
	}

	final void process(final LinkBroken protocol) {
		scheduleSessionTask(protocol.sid, () -> {
			if (Trace.isInfoEnabled())
				Trace.info("process LinkBroken sid = " + protocol.sid);
			removeTransport(protocol.sid);
		});
	}

	final void process(final OnlineAnnounce oa) {
		if (!isStateless())
			scheduleSessionTask(oa.sid, () -> addTransport1(oa));
	}

	final void process(SendControlToServer protocol) {
		if (isStateless())
			statelessAddTransport(protocol.sessionid, protocol.getTransport());
		View view = viewcontext.findView(protocol.sessionid, protocol.classindex, protocol.instanceindex);
		if (null != view) {
			if (protocol.controlindex != -1)
				view.scheduleProcessControl(protocol.controlindex, OctetsStream.wrap(protocol.controlparameter),
						protocol.sessionid);
			else
				view.scheduleOnMessage(protocol.stringdata, protocol.sessionid);
		} else if (Trace.isWarnEnabled())
			Trace.warn("view not found " + protocol);
	}

	private void process(long sessionid, short classindex, int instanceindex, String message) {
		View view = viewcontext.findView(sessionid, classindex, instanceindex);
		if (null != view)
			view.scheduleOnMessage(message, sessionid);
		else if (Trace.isWarnEnabled())
			Trace.warn("view not found classindex " + classindex + " message = " + message);
	}

	private PaySupport checkPaySupport(int payid) {
		PaySupport paySupport = config.getPaySupport();
		if (paySupport == null) {
			if (Trace.isWarnEnabled())
				Trace.warn("PaySupport not defined, but receive payid = " + payid);
			return null;
		}
		if (payid != config.getPayId()) {
			if (Trace.isWarnEnabled())
				Trace.warn("local payid = " + config.getPayId() + " but receive payid = " + payid);
			return null;
		}
		return paySupport;
	}

	final void process(Pay pay) {
		PaySupport paySupport = checkPaySupport(pay.payid);
		scheduleSessionTask(pay.sessionid, () -> {
			try {
				paySupport.accept(pay.serial, pay.sessionid, pay.product, pay.price, pay.count, () -> {
					try {
						new PayAck(pay.payid, pay.serial).send(pay.getTransport());
					} catch (Exception e) {
					}
				});
			} catch (Exception e) {
				if (Trace.isErrorEnabled())
					Trace.error("PaySupport.accept", e);
			}
		});
	}

	final void process(PayAck ack) {
		PaySupport paySupport = checkPaySupport(ack.payid);
		Engine.getApplicationExecutor().execute(() -> {
			try {
				paySupport.cleanup(ack.serial, () -> {
					try {
						new PayAck(ack.payid, ack.serial).send(ack.getTransport());
					} catch (Exception e) {
					}
				});
			} catch (Exception e) {
				if (Trace.isErrorEnabled())
					Trace.error("PaySupport.cleanup", e);
			}
		});
	}

	private interface ToSwitcher {
		void unicastToSwitcher(long sid, int ptype, Octets pdata)
				throws InstantiationException, SizePolicyException, CodecException, ClassCastException;

		void broadcastToSwitcher(int ptype, Octets pdata)
				throws InstantiationException, SizePolicyException, CodecException, ClassCastException;

		void syncViewToClients(SyncViewToClients protocol)
				throws InstantiationException, ClassCastException, SizePolicyException, CodecException;
	}

	private final ToSwitcher createDefaultToSwitcher() {
		return new ToSwitcher() {
			@Override
			public void unicastToSwitcher(long sid, int ptype, Octets pdata)
					throws InstantiationException, SizePolicyException, CodecException, ClassCastException {
				try {
					exchanger.send(new Unicast(sid, ptype, pdata), sid);
				} catch (InstantiationException | SizePolicyException | CodecException | ClassCastException e) {
					throw e;
				} catch (Throwable e) {
					throw new CodecException(e);
				}
			}

			@Override
			public void broadcastToSwitcher(int ptype, Octets pdata)
					throws InstantiationException, SizePolicyException, CodecException, ClassCastException {
				exchanger.send(new Broadcast(ptype, pdata));
			}

			@Override
			public void syncViewToClients(SyncViewToClients protocol)
					throws InstantiationException, ClassCastException, SizePolicyException, CodecException {
				exchanger.send(protocol, protocol.sessionids);
			}
		};
	}

	private volatile ToSwitcher toswitcher = createDefaultToSwitcher();
	private final DirectDispatcher.ProviderDispatchable providerdispatcher = new DirectDispatcher.ProviderDispatchable() {
		private void _dispatchSessionProtocol(final long sid, int ptype, Octets pdata) {
			ProviderTransportImpl transport = transports.get(sid);
			if (null == transport)
				return;
			Skeleton skel;
			try {
				skel = config.getProviderState().decode(ptype & 0xff, pdata, transport);
			} catch (Exception e) {
				if (Trace.isErrorEnabled())
					Trace.error("Provider decode protocol sid = " + sid, e);
				close(transport);
				return;
			}
			try {
				skel.process();
			} catch (Throwable e) {
				if (Trace.isErrorEnabled())
					Trace.error("Provider skel.process() sid = " + sid, e);
				close(transport);
			}
		}

		private void _dispatchWebSocketProtocol(final long sid, Octets pdata) {
			ProviderTransportImpl transport = transports.get(sid);
			if (null == transport)
				return;
			try {
				WebSocketProtocol wsp = new WebSocketProtocol(OctetsStream.wrap(pdata)) {
					@Override
					public void process() throws Exception {
					}
				};
				String s = wsp.getText();
				int p0 = s.indexOf(',') + 1;
				int p1 = s.indexOf(',', p0);
				short classindex = Short.parseShort(s.substring(p0, p1), Character.MAX_RADIX);
				int p2 = s.indexOf(':', ++p1);
				int instanceindex = Integer.parseInt(s.substring(p1, p2), Character.MAX_RADIX);
				String message = s.substring(++p2);
				process(sid, classindex, instanceindex, message);
			} catch (Exception e) {
				if (Trace.isErrorEnabled())
					Trace.error("Provider decode WebSocketProtocol sid = " + sid, e);
				close(transport);
			}
		}

		@Override
		public final void dispatchSessionProtocol(final long sid, final int ptype, final Octets pdata) {
			if (-1 == ptype) {
				_dispatchWebSocketProtocol(sid, pdata);
			} else {
				scheduleSessionTask(sid, () -> _dispatchSessionProtocol(sid, ptype, pdata));
			}
		}

		@Override
		public final void setSwitcherReceiveable(final DirectDispatcher.SwitcherReceiveable recvable) {
			toswitcher = new ToSwitcher() {
				@Override
				public void unicastToSwitcher(long sid, int ptype, Octets pdata) {
					recvable.switcherUnicast(sid, ptype, pdata);
				}

				@Override
				public void broadcastToSwitcher(int ptype, Octets pdata) {
					recvable.switcherBroadcast(ptype, pdata);
				}

				@Override
				public void syncViewToClients(SyncViewToClients protocol)
						throws InstantiationException, ClassCastException, SizePolicyException, CodecException {
					recvable.syncViewToClients(protocol);
				}
			};
		}

		@Override
		public void onViewControl(final SendControlToServer protocol) {
			process(protocol);
		}
	};

	@Override
	public Manager getWrapperManager() {
		return wrapper;
	}

	void check(int type, int size) throws InstantiationException, SizePolicyException {
		config.getProviderState().check(type & 0xff, size);
	}

	Manager getOutmostWrapperManager() {
		Manager manager = this;
		while (manager.getWrapperManager() != null)
			manager = manager.getWrapperManager();
		return manager;
	}
}
