package limax.zdb;

import java.util.Iterator;
import java.util.Map;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import limax.codec.MarshalException;
import limax.codec.Octets;
import limax.codec.OctetsStream;

final class TStorage<K, V> implements StorageInterface {
	private StorageEngine engine;

	TStorage(LoggerEngine logger, String tableName) {
		switch (Zdb.meta().getEngineType()) {
		case MYSQL:
			engine = new StorageMysql((LoggerMysql) logger, tableName);
			break;
		case EDB:
			engine = new StorageEdb((LoggerEdb) logger, tableName);
			break;
		default:
			throw new XError("unknown engine type");
		}
	}

	@Override
	public StorageEngine getEngine() {
		return engine;
	}

	private ConcurrentMap<K, TRecord<K, V>> changed = new ConcurrentHashMap<>();
	private ConcurrentMap<K, TRecord<K, V>> marshal = new ConcurrentHashMap<>();
	private ConcurrentMap<K, TRecord<K, V>> snapshot = new ConcurrentHashMap<>();
	private final ReentrantReadWriteLock snapshotClearLock = new ReentrantReadWriteLock();

	// from TTable
	void onRecordChange(TRecord<K, V> r) {
		if (Transaction.current().duration() != null) {
			r.marshalA();
			if (r.getState() == TRecord.State.REMOVE) {
				marshal.remove(r.getKey());
			} else {
				marshal.put(r.getKey(), r);
			}
		} else {
			if (r.getState() == TRecord.State.REMOVE) {
				changed.remove(r.getKey());
				marshal.remove(r.getKey());
			} else {
				changed.put(r.getKey(), r);
			}
		}
	}

	// from TTableCache.Cleaner
	boolean isChangedOrUnknown(K key) {
		return changed.containsKey(key) || marshal.containsKey(key);
	}

	private volatile long countMarshalN = 0;
	private volatile long countMarshal0 = 0;
	private volatile long countFlush = 0;
	private volatile long countSnapshot = 0;
	private volatile long countMarshalNTryFail = 0;
	volatile long flushKeySize = 0;
	volatile long flushValueSize = 0;

	@Override
	public int marshalN() {
		int marshaled = 0;
		int tryFail = 0;
		for (Iterator<TRecord<K, V>> it = changed.values().iterator(); it.hasNext();) {
			TRecord<K, V> r = it.next();
			if (r.tryMarshalN()) {
				marshal.put(r.getKey(), r);
				it.remove();
				++marshaled;
			} else
				++tryFail;
		}
		this.countMarshalN += marshaled;
		this.countMarshalNTryFail += tryFail;
		return marshaled;
	}

	@Override
	public int marshal0() {
		marshal.putAll(changed);
		changed.values().forEach(TRecord::marshal0);
		int size = changed.size();
		changed.clear();
		this.countMarshal0 += size;
		return size;
	}

	@Override
	public int snapshot() {
		final ConcurrentMap<K, TRecord<K, V>> tmp = this.snapshot;
		this.snapshot = this.marshal;
		this.marshal = tmp;
		snapshot.values().forEach(TRecord::snapshot);
		this.countSnapshot += snapshot.size();
		return snapshot.size();
	}

	@Override
	public int flush() {
		int count = (int) snapshot.values().stream().filter(r -> r.flush(this)).count();
		final Map<K, TRecord<K, V>> tmp;
		final Lock writeLock = snapshotClearLock.writeLock();
		writeLock.lock();
		try {
			tmp = this.snapshot;
			this.snapshot = new ConcurrentHashMap<>();
		} finally {
			writeLock.unlock();
		}
		tmp.values().forEach(r -> r.clear());
		this.countFlush += count;
		return count;
	}

	final boolean exist(K key, TTable<K, V> table) {
		final Lock readLock = snapshotClearLock.readLock();
		readLock.lock();
		try {
			TRecord<K, V> r = snapshot.get(key);
			if (null != r)
				return r.exist();
		} finally {
			readLock.unlock();
		}
		return engine.exist(table.marshalKey(key));
	}

	V find(K key, TTable<K, V> table) {
		Octets value = null;
		boolean foundInSnapshot = false;
		final Lock readLock = snapshotClearLock.readLock();
		readLock.lock();
		try {
			TRecord<K, V> r = snapshot.get(key);
			if (null != r) {
				foundInSnapshot = true;
				value = r.find();
			}
		} finally {
			readLock.unlock();
		}
		if (!foundInSnapshot)
			value = engine.find(table.marshalKey(key));
		try {
			return value != null ? table.unmarshalValue(OctetsStream.wrap(value)) : null;
		} catch (MarshalException e) {
			throw new RuntimeException(e);
		}
	}

	public long getCountFlush() {
		return countFlush;
	}

	public long getCountMarshal0() {
		return countMarshal0;
	}

	public long getCountMarshalN() {
		return countMarshalN;
	}

	public long getCountMarshalNTryFail() {
		return countMarshalNTryFail;
	}

	public long getCountSnapshot() {
		return countSnapshot;
	}

	public long getFlushKeySize() {
		return flushKeySize;
	}

	public long getFlushValueSize() {
		return flushValueSize;
	}
}
